/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.api;

import org.cementframework.querybyproxy.shared.api.model.QueryFragment;
import org.cementframework.querybyproxy.shared.api.model.values.QueryValue;
import org.cementframework.querybyproxy.shared.api.model.values.Subquery;

/**
 * The primary entry point into creating by-proxy criteria queries.
 *
 * <p>Represent query that allows only strictly-typed grammar as parameters
 * for its methods.</p>
 *
 * @author allenparslow
 *
 * @param <T>
 *            the root type of the query (may be effectively modified by calling
 *            select(proxy)).
 */
public interface StrictQuery<T> extends QueryFragment {

    /**
     * Inner joins a recorded entity-proxy nested-property method call to the
     * current from-clause.
     *
     * @param <J>
     *            type-inference.
     * @param targetProperty
     *            a recorded entity-proxy method call.
     * @return A entity-proxy to the nested property.
     */
    <J> J join(QueryValue<J> targetProperty);

    /**
     * Left joins a recorded entity-proxy nested-property method call to the
     * current from-clause.
     *
     * @param <J>
     *            type-inference.
     * @param targetProperty
     *            a recorded entity-proxy method call.
     * @return A entity-proxy to the nested property.
     */
    <J> J leftJoin(QueryValue<J> targetProperty);

    /**
     * Right joins a recorded entity-proxy nested-property method call to the
     * current from-clause.
     *
     * @param <J>
     *            type-inference.
     * @param targetProperty
     *            a recorded entity-proxy method call.
     * @return A entity-proxy to the nested property.
     */
    <J> J rightJoin(QueryValue<J> targetProperty);

    /**
     * Fetch joins a recorded entity-proxy nested-property method call to the
     * current from-clause.
     *
     * @param <J>
     *            type-inference.
     * @param targetProperty
     *            a recorded entity-proxy method call.
     * @return A entity-proxy to the nested property.
     */
    <J> J joinFetch(QueryValue<J> targetProperty);

    /**
     * Selects a singular value.
     *
     * @param <S>
     *            type inference based on <code>QueryValue</code>.
     *
     * @param value
     *            selection for the query.
     *
     * @return a typed query used to find results.
     */
    <S> TypedQuery<S> select(QueryValue<S> value);

    /**
     * Create a sub-query instance using the current query.
     *
     * @param <S>
     *            The type of sub-query given the specified
     *            <code>QueryValue</code>
     * @param value
     *            the <code>QueryValue</code> representing the singular return
     *            type of the subquery.
     * @return a subquery (immutable).
     */
    <S> Subquery<S> subquery(QueryValue<S> value);
}
