/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.api;

import org.cementframework.querybyproxy.shared.api.model.conditionals.Conditional;
import org.cementframework.querybyproxy.shared.api.model.sorts.DirectionalQuerySort;
import org.cementframework.querybyproxy.shared.api.model.values.SimpleQueryValue;
import org.cementframework.querybyproxy.shared.api.model.values.StrictQueryValue;
import org.cementframework.querybyproxy.shared.api.model.values.Subquery;

/**
 * Represent query builder that allows only strictly-typed grammar as parameters
 * for its methods.
 *
 * @author allenparslow
 */
public interface StrictQueryBuilder {

    /**
     * Creates proxy for the specified class.
     *
     * @param <T>
     *            type-inference.
     * @param type
     *            the type to proxy.
     * @return the new proxy.
     */
    <T> T createProxy(Class<T> type);

    /**
     * Gets a query-value expression using the last recorded proxy method call.
     *
     * @param <T>
     *            type-inference.
     * @param proxyMethodCall
     *            a recorded proxy method call.
     * @return a query-value expression.
     */
    <T> StrictQueryValue<T> get(T proxyMethodCall);

    /**
     * Gets a query-value expression with an overridden type using the last
     * recorded proxy method call.
     *
     * <p>
     * Commonly used when dealing with aliased properties as the alias type
     * might not be of the same type as the recorded method call's type
     * </p>
     *
     * @param <T>
     *            type-inference using the specified returnType.
     * @param proxyMethodCall
     *            a recorded proxy method call.
     * @param returnType
     *            The overridden return type.
     * @return a query-value expression.
     */
    <T> StrictQueryValue<T> get(Object proxyMethodCall, Class<T> returnType);

    /**
     * Produces a literal (non-parameter) expression using the passed object.
     *
     * <p>
     * SQL Injection Warning: Do NOT use with end-user specified text as it is
     * treated as a literal string during query compilation
     * </p>
     *
     * @param <T>
     *            type-inference.
     * @param object
     *            the target object to be used as a literal.
     * @return a query-value expression.
     */
    <T> StrictQueryValue<T> literal(T object);

    /**
     * Produces a jpql bindable for the target object (produces ":paramX" syntax
     * during query compilation).
     *
     * @param <T>
     *            type-inference.
     * @param target
     *            the target object to be used a bindable query-parameter.
     * @return a query-value expression.
     */
    <T> StrictQueryValue<T> param(T target);

    /**
     * Used to specify that the next conditional-expression should be created as
     * an or-expression, overriding the default conjunction/logic-gate (which is
     * <b>and</b>).
     *
     * @return a conditional-expression placeholder.
     */
    Conditional or();

    /**
     * Creates a group "(...)" wrapper around the specified conditional
     * expressions.
     *
     * @param conditionalExpressions
     *            the conditional-expressions to group.
     * @return the group wrapper.
     */
    Conditional group(Conditional... conditionalExpressions);

    /**
     * Creates a query descending sort using the specified selection.
     *
     * @param selection
     *            a selection.
     * @return a descending sort
     */
    @SuppressWarnings("unchecked")
    DirectionalQuerySort desc(SimpleQueryValue selection);

    /**
     * Creates a "count(1)" expression.
     *
     * @return a query-value expression.
     */
    StrictQueryValue<Long> count();

    /**
     * Create a "count" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    <T> StrictQueryValue<T> count(SimpleQueryValue<T> selection);

    /**
     * Create a "min" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    <T> StrictQueryValue<T> min(SimpleQueryValue<T> selection);

    /**
     * Create a "max" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    <T> StrictQueryValue<T> max(SimpleQueryValue<T> selection);

    /**
     * Create a "sum" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    <T> StrictQueryValue<T> sum(SimpleQueryValue<T> selection);

    /**
     * Create a "count" aggregate using the specified recorded proxy method
     * call.
     *
     * @param proxyMethodCall
     *            a recorded proxy method call. a recorded proxy method call.
     * @return a query-value expression.
     */
    StrictQueryValue<Long> count(Object proxyMethodCall);

    /**
     * Create a "avg" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    <T> StrictQueryValue<T> avg(SimpleQueryValue<T> selection);

    /**
     * Create a "size" modifier using the specified selection.
     *
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    @SuppressWarnings("unchecked")
    StrictQueryValue<Long> size(SimpleQueryValue selection);

    /**
     * Creates a "exists" subquery modifier around a subquery.
     *
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper conditional-expression.
     */
    @SuppressWarnings("unchecked")
    Conditional exists(Subquery subquery);

    /**
     * Creates a "not exists" subquery modifier around a subquery.
     *
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper conditional-expression.
     */
    @SuppressWarnings("unchecked")
    Conditional notExists(Subquery subquery);

    /**
     * Creates a "all" subquery modifier around a subquery.
     *
     * @param <T>
     *            type-inference.
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper.
     */
    <T> Subquery<T> all(Subquery<T> subquery);

    /**
     * Creates a "any" subquery modifier around a subquery.
     *
     * @param <T>
     *            type-inference.
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper.
     */
    <T> Subquery<T> any(Subquery<T> subquery);

    /**
     * Creates a "some" subquery modifier around a subquery.
     *
     * @param <T>
     *            type-inference.
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper.
     */
    <T> Subquery<T> some(Subquery<T> subquery);

    /**
     * Creates a query value expression used either a recorded method call, a
     * literal value or query parameter.
     *
     * @param <T>
     *            type-inference.
     * @param target
     *            the target value or recorded method call.
     * @return the query value expression.
     */
    <T> StrictQueryValue<T> createQueryValue(T target);

    /**
     * Creates a query value expression used either a recorded method call, a
     * literal value or query parameter.
     *
     * @param <T>
     *            type-inference.
     * @param target
     *            the target value or recorded method call.
     * @param useFirstCall
     *            True, if the first call on the recording stack should be used.
     * @return the query value expression.
     */
    <T> StrictQueryValue<T> createQueryValue(T target, boolean useFirstCall);
}
