/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.api.model;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

import org.cementframework.querybyproxy.shared.api.model.joins.AbstractJoin;
import org.cementframework.querybyproxy.shared.api.model.joins.QueryJoin;
import org.cementframework.querybyproxy.shared.api.model.joins.QueryJoinModifier;
import org.cementframework.querybyproxy.shared.api.model.joins.QueryJoinType;

/**
 * Represents a "from" clause.
 *
 * @author allenparslow
 */
public class FromClause extends AbstractJoin {

    private final String name;

    private final Map<Object, QueryJoin> joinMap;

    /**
     * Create a new <code>FromImpl</code> instance.
     *
     * @param rootName
     *            the name of the root entity.
     * @param rootProxy
     *            the root-proxy
     */
    public FromClause(String rootName, Object rootProxy) {
        super(null, QueryJoinType.INNER, QueryJoinModifier.NONE, rootProxy);
        this.name = rootName;

        joinMap = Collections.unmodifiableMap(new LinkedHashMap<Object, QueryJoin>());
    }

    /**
     * Create a new <code>FromImpl</code> instance.
     *
     * <p>
     * <code>FromImpl</code> is immutable, this constructor is used to clone a
     * from, adding a new join
     * </p>
     *
     * @param join
     *            the join to add.
     * @param baseFrom
     *            the
     */
    public FromClause(QueryJoin join, FromClause baseFrom) {
        super(null,
                QueryJoinType.INNER,
                QueryJoinModifier.NONE,
                baseFrom.getTargetProxy());
        this.name = baseFrom.getRootName();

        Map<Object, QueryJoin> newMap = new LinkedHashMap<Object, QueryJoin>(baseFrom.getJoinMap());
        newMap.put(join.getTargetProxy(), join);

        joinMap = Collections.unmodifiableMap(newMap);
    }

    /**
     * Gets the name of the root entity.
     *
     * @return the root entity's name.
     */
    public String getRootName() {
        return name;
    }

    /**
     * Gets the map of entity-proxies to their respective joins.
     *
     * @return the entity-proxy join map.
     */
    public Map<Object, QueryJoin> getJoinMap() {
        return joinMap;
    }

    /**
     * Finds the parent join for the specified entity proxy.
     *
     * @param parent
     *            the proxy to query.
     *
     * @return The parent proxy's join.
     */
    public QueryJoin findParentJoin(Object parent) {
        if (this.getTargetProxy().equals(parent)) {
            return this;
        }
        return joinMap.get(parent);
    }

}
