/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.api.model.values;

import static org.cementframework.querybyproxy.shared.impl.StaticProxyQueryBuilder.*;

import java.util.ArrayList;
import java.util.List;

import org.cementframework.querybyproxy.shared.api.model.conditionals.BinaryConditional;
import org.cementframework.querybyproxy.shared.api.model.conditionals.ComparisonOperator;
import org.cementframework.querybyproxy.shared.api.model.conditionals.UnaryConditional;

/**
 * Abstract representation a query-value.
 *
 * @author allenparslow
 * @param <T>
 *            type-inference
 */
public abstract class AbstractQueryValue<T> extends AbstractStrictQueryValue<T>
        implements QueryValue<T> {

    /**
     * {@inheritDoc}
     */
    public UnaryConditional<T> isNotEmpty() {
        return unaryOperation(ComparisonOperator.IS_NOT_EMPTY);
    }

    /**
     * {@inheritDoc}
     */
    public UnaryConditional<T> isEmpty() {
        return unaryOperation(ComparisonOperator.IS_EMPTY);
    }

    /**
     * {@inheritDoc}
     */
    public UnaryConditional<T> isNull() {
        return unaryOperation(ComparisonOperator.IS_NULL);
    }

    /**
     * {@inheritDoc}
     */
    public UnaryConditional<T> isNotNull() {
        return unaryOperation(ComparisonOperator.IS_NOT_NULL);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> notIn(T... values) {
        List<QueryLiteral<T>> params = new ArrayList<QueryLiteral<T>>(values.length);
        for (T value : values) {
            params.add(new QueryLiteral<T>(value));
        }

        return inOperator(params, ComparisonOperator.NOT_IN);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> in(T... values) {
        List<QueryLiteral<T>> params = new ArrayList<QueryLiteral<T>>(values.length);
        for (T value : values) {
            params.add(new QueryLiteral<T>(value));
        }

        return inOperator(params, ComparisonOperator.IN);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> notLike(T value) {
        return binaryOperation(createQueryValue(value),
                ComparisonOperator.NOT_LIKE);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> like(T value) {
        return binaryOperation(createQueryValue(value), ComparisonOperator.LIKE);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> equalTo(T value) {
        return binaryOperation(createQueryValue(value),
                ComparisonOperator.EQUALS);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> notEqualTo(T value) {
        return binaryOperation(createQueryValue(value),
                ComparisonOperator.NOT_EQUALS);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> greaterThan(T value) {
        return binaryOperation(createQueryValue(value),
                ComparisonOperator.GREATER_THAN);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> greaterThanOrEqualTo(T value) {
        return binaryOperation(createQueryValue(value),
                ComparisonOperator.GREATER_THAN_OR_EQUAL_TO);

    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> lessThan(T value) {
        return binaryOperation(createQueryValue(value),
                ComparisonOperator.LESS_THAN);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> lessThanOrEqualTo(T value) {
        return binaryOperation(createQueryValue(value),
                ComparisonOperator.LESS_THAN_OR_EQUAL_TO);
    }

}

