/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.api.model.values;

import java.util.List;

import org.cementframework.querybyproxy.shared.api.ProxyQuerySessions;
import org.cementframework.querybyproxy.shared.api.model.conditionals.BinaryConditional;
import org.cementframework.querybyproxy.shared.api.model.conditionals.ComparisonOperator;
import org.cementframework.querybyproxy.shared.api.model.conditionals.LogicGate;
import org.cementframework.querybyproxy.shared.api.model.conditionals.UnaryConditional;

/**
 * An abstract representation a value or path expression used in a query.
 *
 * <p>
 * Contains a subset of possible operation (all very strictly typed) that can be
 * emitted from the value.
 * </p>
 *
 * @author allenparslow
 * @param <T>
 */
public class AbstractStrictQueryValue<T> implements StrictQueryValue<T> {

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> notLike(QueryValue<T> value) {
        return binaryOperation(value, ComparisonOperator.NOT_LIKE);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> like(QueryValue<T> value) {
        return binaryOperation(value, ComparisonOperator.LIKE);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> notIn(List<? extends QueryValue<T>> values) {

        return inOperator(values, ComparisonOperator.NOT_IN);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> in(List<? extends QueryValue<T>> values) {

        return inOperator(values, ComparisonOperator.IN);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public BinaryConditional<T> memberOf(QueryValue<List<T>> value) {
        QueryValue<T> expValue = (QueryValue<T>) value;
        BinaryConditional exp = binaryOperation(expValue, ComparisonOperator.MEMBER_OF);
        return (BinaryConditional<T>) exp;
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public BinaryConditional<T> notMemberOf(QueryValue<List<T>> value) {
        QueryValue<T> expValue = (QueryValue<T>) value;
        BinaryConditional exp = binaryOperation(expValue, ComparisonOperator.NOT_MEMBER_OF);
        return (BinaryConditional<T>) exp;
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> equalTo(QueryValue<T> value) {
        return binaryOperation(value, ComparisonOperator.EQUALS);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> notEqualTo(QueryValue<T> value) {
        return binaryOperation(value, ComparisonOperator.NOT_EQUALS);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> greaterThan(QueryValue<T> value) {
        return binaryOperation(value, ComparisonOperator.GREATER_THAN);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> greaterThanOrEqualTo(QueryValue<T> value) {
        return binaryOperation(value, ComparisonOperator.GREATER_THAN_OR_EQUAL_TO);

    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> lessThan(QueryValue<T> value) {
        return binaryOperation(value, ComparisonOperator.LESS_THAN);
    }

    /**
     * {@inheritDoc}
     */
    public BinaryConditional<T> lessThanOrEqualTo(QueryValue<T> value) {
        return binaryOperation(value, ComparisonOperator.LESS_THAN_OR_EQUAL_TO);
    }

    protected UnaryConditional<T> unaryOperation(ComparisonOperator expression) {
        LogicGate logicGate = ProxyQuerySessions.get().getNextLogicGate();

        UnaryConditional<T> conditional = new UnaryConditional<T>(
                logicGate, this, expression);
        return conditional;
    }

    protected BinaryConditional<T> binaryOperation(
            QueryValue<T> value,
            ComparisonOperator expression) {

        LogicGate logicGate = ProxyQuerySessions.get().getNextLogicGate();

        if (value == null) {
            throw new IllegalArgumentException(
                    "Binary expression query-value must not be null: "
                    + " check that querybuilder.get() was called"
                    + " OR unary isNull/isNotNull was intended");
        }
        BinaryConditional<T> conditional = new BinaryConditional<T>(
                logicGate, this, expression, value);

        return conditional;
    }

    protected BinaryConditional<T> inOperator(List<? extends QueryValue<T>> params,
            ComparisonOperator expression) {

        return binaryOperation(
                new ValuesList<T>(params),
                expression);
    }
}
