/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.api.model.values;

import java.util.List;

import org.cementframework.querybyproxy.shared.api.model.conditionals.BinaryConditional;
import org.cementframework.querybyproxy.shared.api.model.conditionals.UnaryConditional;

/**
 * Represents a value or path expression used in a query.
 *
 * <p>Contains the possible operation that can be emitted from the value.</p>
 *
 * @author allenparslow
 *
 * @param <T>
 *            the type of value.
 */
public interface QueryValue<T> extends StrictQueryValue<T> {

    /**
     * Creates a "=" binary expression.
     *
     * @param value
     *            a value or recorded method call
     * @return the binary expression.
     */
    BinaryConditional<T> equalTo(T value);

    /**
     * Creates a "!=" binary expression.
     *
     * @param value
     *            a value or recorded method call
     * @return the binary expression.
     */
    BinaryConditional<T> notEqualTo(T value);

    /**
     * Creates a ">" binary expression.
     *
     * @param value
     *            a value or recorded method call
     * @return the binary expression.
     */
    BinaryConditional<T> greaterThan(T value);

    /**
     * Creates a ">=" binary expression.
     *
     * @param value
     *            a value or recorded method call
     * @return the binary expression.
     */
    BinaryConditional<T> greaterThanOrEqualTo(T value);

    /**
     * Creates a "<" binary expression.
     *
     * @param value
     *            a value or recorded method call
     * @return the binary expression.
     */
    BinaryConditional<T> lessThan(T value);

    /**
     * Creates a "<=" binary expression.
     *
     * @param value
     *            a value or recorded method call
     * @return the binary expression.
     */
    BinaryConditional<T> lessThanOrEqualTo(T value);

    /**
     * Creates a "like" binary expression.
     *
     * @param value
     *            a value or recorded method call
     * @return the binary expression.
     */
    BinaryConditional<T> like(T value);

    /**
     * Creates a "like" binary expression.
     *
     * @param value
     *            a <code>QueryValue</code>.
     * @return the binary expression.
     */
    BinaryConditional<T> like(QueryValue<T> value);

    /**
     * Creates a "not like" binary expression.
     *
     * @param value
     *            a value or recorded method call
     * @return the binary expression.
     */
    BinaryConditional<T> notLike(T value);

    /**
     * Creates a "not like" binary expression.
     *
     * @param value
     *            a <code>QueryValue</code>.
     * @return the binary expression.
     */
    BinaryConditional<T> notLike(QueryValue<T> value);

    /**
     * Creates a "in" binary expression.
     *
     * @param values
     *            var-arg of values or recorded method calls
     * @return the binary expression.
     */
    BinaryConditional<T> in(T... values);

    /**
     * Creates a "in" binary expression.
     *
     * @param values
     *            a list of <code>QueryValue</code>.
     * @return the binary expression.
     */
    BinaryConditional<T> in(List<? extends QueryValue<T>> values);

    /**
     * Creates a "not in" binary expression.
     *
     * @param values
     *            var-arg of values or recorded method calls
     * @return the binary expression.
     */
    BinaryConditional<T> notIn(T... values);

    /**
     * Creates a "not in" binary expression.
     *
     * @param values
     *            a list of <code>QueryValue</code>.
     * @return the binary expression.
     */
    BinaryConditional<T> notIn(List<? extends QueryValue<T>> values);

    /**
     * Creates a "is empty" unary expression.
     *
     * @return the unary expression.
     */
    UnaryConditional<T> isEmpty();

    /**
     * Creates a "is not empty" unary expression.
     *
     * @return the unary expression.
     */
    UnaryConditional<T> isNotEmpty();

    /**
     * Creates a "is null" unary expression.
     *
     * @return the unary expression.
     */
    UnaryConditional<T> isNull();

    /**
     * Creates a "is not null" unary expression.
     *
     * @return the unary expression.
     */
    UnaryConditional<T> isNotNull();

    /**
     * Creates a "member of" binary expression.
     *
     * @param value
     *            a qb.get(recorded method call) of a list property.
     * @return the binary expression.
     */
    BinaryConditional<T> memberOf(QueryValue<List<T>> value);

    /**
     * Creates a "not member of" binary expression.
     *
     * @param value
     *            a qb.get(recorded method call) of a list property.
     * @return the binary expression.
     */
    BinaryConditional<T> notMemberOf(QueryValue<List<T>> value);
}
