/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.api.model.values;

import java.util.List;

/**
 * A query-value that re-directs its methods to a callback.
 *
 * @author allenparslow
 * @param <T>
 *            type-inference
 */
public class RedirectingQueryValue<T> {
    private final QueryValue<T> delegate;
    private final ConditionalExpressionCallback callback;

    /**
     * Create a new <code>QueryValueDelegate</code> instance.
     *
     * @param delegate
     *            the query-value to wrap
     * @param callback
     *            the callback to redirect method call results to.
     */
    public RedirectingQueryValue(QueryValue<T> delegate, ConditionalExpressionCallback callback) {
        this.delegate = delegate;
        this.callback = callback;
    }

    /**
     * Creates a "=" binary expression.
     *
     * @param value
     *            a value or recorded method call
     */
    public void equalTo(T value) {
        callback.processConditional(delegate.equalTo(value));
    }

    /**
     * Creates a "=" binary expression.
     *
     * @param value
     *            a query value.
     */
    public void equalTo(QueryValue<T> value) {
        callback.processConditional(delegate.equalTo(value));
    }

    /**
     * Creates a ">" binary expression.
     *
     * @param value
     *            a value or recorded method call
     */
    public void greaterThan(T value) {
        callback.processConditional(delegate.greaterThan(value));
    }

    /**
     * Creates a ">" binary expression.
     *
     * @param value
     *            a query value.
     */
    public void greaterThan(QueryValue<T> value) {
        callback.processConditional(delegate.greaterThan(value));
    }

    /**
     * Creates a ">=" binary expression.
     *
     * @param value
     *            a value or recorded method call
     */
    public void greaterThanOrEqualTo(T value) {
        callback.processConditional(delegate.greaterThanOrEqualTo(value));
    }

    /**
     * Creates a ">=" binary expression.
     *
     * @param value
     *            a query value.
     */
    public void greaterThanOrEqualTo(QueryValue<T> value) {
        callback.processConditional(delegate.greaterThanOrEqualTo(value));
    }

    /**
     * Creates a "in" binary expression.
     *
     * @param values
     *            var-arg of values or recorded method calls
     */
    public void in(T... values) {
        callback.processConditional(delegate.in(values));
    }

    /**
     * Creates a "in" binary expression.
     *
     * @param values
     *            a list of values to use in the expression.
     */
    public void in(List<? extends QueryValue<T>> values) {
        callback.processConditional(delegate.in(values));
    }

    /**
     * Creates a "is empty" unary expression.
     */
    public void isEmpty() {
        callback.processConditional(delegate.isEmpty());
    }

    /**
     * Creates a "is not empty" unary expression.
     */
    public void isNotEmpty() {
        callback.processConditional(delegate.isNotEmpty());
    }

    /**
     * Creates a "is not null" unary expression.
     */
    public void isNotNull() {
        callback.processConditional(delegate.isNotNull());
    }

    /**
     * Creates a "is null" unary expression.
     */
    public void isNull() {
        callback.processConditional(delegate.isNull());
    }

    /**
     * Creates a "<" binary expression.
     *
     * @param value
     *            a value or recorded method call
     */
    public void lessThan(T value) {
        callback.processConditional(delegate.lessThan(value));
    }

    /**
     * Creates a "<" binary expression.
     *
     * @param value
     *            a query value.
     */
    public void lessThan(QueryValue<T> value) {
        callback.processConditional(delegate.lessThan(value));
    }

    /**
     * Creates a "<=" binary expression.
     *
     * @param value
     *            a value or recorded method call
     */
    public void lessThanOrEqualTo(T value) {
        callback.processConditional(delegate.lessThanOrEqualTo(value));
    }

    /**
     * Creates a "<=" binary expression.
     *
     * @param value
     *            a query value.
     */
    public void lessThanOrEqualTo(QueryValue<T> value) {
        callback.processConditional(delegate.lessThanOrEqualTo(value));
    }

    /**
     * Creates a "like" binary expression.
     *
     * @param value
     *            a value or recorded method call
     */
    public void like(T value) {
        callback.processConditional(delegate.like(value));
    }

    /**
     * Creates a "like" binary expression.
     *
     * @param value
     *            a query value.
     */
    public void like(QueryValue<T> value) {
        callback.processConditional(delegate.like(value));
    }

    /**
     * Creates a "member of" binary expression.
     *
     * @param value
     *            A query value representing a list of values.
     */
    public void memberOf(QueryValue<List<T>> value) {
        callback.processConditional(delegate.memberOf(value));
    }

    /**
     * Creates a "!=" binary expression.
     *
     * @param value
     *            a value or recorded method call
     */
    public void notEqualTo(T value) {
        callback.processConditional(delegate.notEqualTo(value));
    }

    /**
     * Creates a "!=" binary expression.
     *
     * @param value
     *            a query value.
     */
    public void notEqualTo(QueryValue<T> value) {
        callback.processConditional(delegate.notEqualTo(value));
    }

    /**
     * Creates a "not in" binary expression.
     *
     * @param values
     *            var-arg of values or recorded method calls
     */
    public void notIn(T... values) {
        callback.processConditional(delegate.notIn(values));
    }

    /**
     * Creates a "not in" binary expression.
     *
     * @param values
     *            a list of values to use in the expression.
     */
    public void notIn(List<? extends QueryValue<T>> values) {
        callback.processConditional(delegate.notIn(values));
    }

    /**
     * Creates a "not like" binary expression.
     *
     * @param value
     *            a value or recorded method call
     */
    public void notLike(T value) {
        callback.processConditional(delegate.notLike(value));
    }

    /**
     * Creates a "not like" binary expression.
     *
     * @param value
     *            a query value.
     */
    public void notLike(QueryValue<T> value) {
        callback.processConditional(delegate.notLike(value));
    }

    /**
     * Creates a "not member of" binary expression.
     *
     * @param value
     *            A query value representing a list of values.
     */
    public void notMemberOf(QueryValue<List<T>> value) {
        callback.processConditional(delegate.notMemberOf(value));
    }
}
