/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.api.model.values;

import java.util.List;

import org.cementframework.querybyproxy.shared.api.model.conditionals.BinaryConditional;

/**
 * Represents a value or path expression used in a query.
 *
 * <p>
 * Contains a subset of possible operation (all very strictly typed) that can be
 * emitted from the value.
 * </p>
 *
 * @author allenparslow
 * @param <T>
 */
public interface StrictQueryValue<T> extends SimpleQueryValue<T> {

    /**
     * Creates a "=" binary expression.
     *
     * @param value
     *            a query value.
     * @return the binary expression.
     */
    BinaryConditional<T> equalTo(QueryValue<T> value);

    /**
     * Creates a "!=" binary expression.
     *
     * @param value
     *            a query value.
     * @return the binary expression.
     */
    BinaryConditional<T> notEqualTo(QueryValue<T> value);

    /**
     * Creates a ">" binary expression.
     *
     * @param value
     *            a query value.
     * @return the binary expression.
     */
    BinaryConditional<T> greaterThan(QueryValue<T> value);

    /**
     * Creates a ">=" binary expression.
     *
     * @param value
     *            a query value.
     * @return the binary expression.
     */
    BinaryConditional<T> greaterThanOrEqualTo(QueryValue<T> value);

    /**
     * Creates a "<" binary expression.
     *
     * @param value
     *            a query value.
     * @return the binary expression.
     */
    BinaryConditional<T> lessThan(QueryValue<T> value);

    /**
     * Creates a "<=" binary expression.
     *
     * @param value
     *            a query value.
     * @return the binary expression.
     */
    BinaryConditional<T> lessThanOrEqualTo(QueryValue<T> value);

    /**
     * Creates a "in" binary expression.
     *
     * @param values a list of values to use in the expression.
     * @return the binary expression.
     */
    BinaryConditional<T> in(List<? extends QueryValue<T>> values);

    /**
     * Creates a "not in" binary expression.
     *
     * @param values a list of values to use in the expression.
     * @return the binary expression.
     */
    BinaryConditional<T> notIn(List<? extends QueryValue<T>> values);

    /**
     * Creates a "member of" binary expression.
     *
     * @param value A query value representing a list of values.
     * @return the binary expression.
     */
    BinaryConditional<T> memberOf(QueryValue<List<T>> value);

    /**
     * Creates a "not member of" binary expression.
     *
     * @param value A query value representing a list of values.
     * @return the binary expression.
     */
    BinaryConditional<T> notMemberOf(QueryValue<List<T>> value);

    /**
     * Creates a "like" binary expression.
     *
     * @param value a query value.
     * @return the binary expression.
     */
    BinaryConditional<T> like(QueryValue<T> value);

    /**
     * Creates a "not like" binary expression.
     *
     * @param value a query value.
     * @return the binary expression.
     */
    BinaryConditional<T> notLike(QueryValue<T> value);
}
