/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.impl;

import static org.cementframework.querybyproxy.shared.impl.StaticProxyQueryBuilder.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.cementframework.querybyproxy.shared.api.DynamicQuery;
import org.cementframework.querybyproxy.shared.api.ProxyQuery;
import org.cementframework.querybyproxy.shared.api.TypedQuery;
import org.cementframework.querybyproxy.shared.api.model.FromClause;
import org.cementframework.querybyproxy.shared.api.model.GroupByClause;
import org.cementframework.querybyproxy.shared.api.model.HavingClause;
import org.cementframework.querybyproxy.shared.api.model.OrderByClause;
import org.cementframework.querybyproxy.shared.api.model.SelectClause;
import org.cementframework.querybyproxy.shared.api.model.WhereClause;
import org.cementframework.querybyproxy.shared.api.model.conditionals.Conditional;
import org.cementframework.querybyproxy.shared.api.model.joins.NestedPropertyJoin;
import org.cementframework.querybyproxy.shared.api.model.joins.QueryJoin;
import org.cementframework.querybyproxy.shared.api.model.joins.QueryJoinModifier;
import org.cementframework.querybyproxy.shared.api.model.joins.QueryJoinType;
import org.cementframework.querybyproxy.shared.api.model.joins.ThetaJoin;
import org.cementframework.querybyproxy.shared.api.model.selections.ConstructorValue;
import org.cementframework.querybyproxy.shared.api.model.selections.SelectProxy;
import org.cementframework.querybyproxy.shared.api.model.selections.Selection;
import org.cementframework.querybyproxy.shared.api.model.sorts.DirectionalQuerySort;
import org.cementframework.querybyproxy.shared.api.model.sorts.QuerySort;
import org.cementframework.querybyproxy.shared.api.model.sorts.QuerySortOperator;
import org.cementframework.querybyproxy.shared.api.model.values.ConditionalExpressionCallback;
import org.cementframework.querybyproxy.shared.api.model.values.ProxyPathExpression;
import org.cementframework.querybyproxy.shared.api.model.values.QueryValue;
import org.cementframework.querybyproxy.shared.api.model.values.RedirectingQueryValue;
import org.cementframework.querybyproxy.shared.api.model.values.Subquery;
import org.cementframework.querybyproxy.shared.api.model.values.SubqueryValue;
import org.cementframework.recordingproxy.api.RecordedMethodCall;
import org.cementframework.recordingproxy.api.RecordingSessions;
import org.cementframework.recordingproxy.impl.MethodCallUtils;

/**
 * The primary entry point into creating by-proxy criteria queries.
 *
 * @author allenparslow
 * @param <T>
 *            the root type of the query (may be effectively modified by calling
 *            select(proxy)).
 */
public abstract class AbstractProxyQueryImpl<T>
        implements ProxyQuery<T>, ConditionalExpressionCallback {

    private final Class<?> rootProxyClass;
    private final T proxy;

    private DynamicQuery query;

    private SelectClause select;
    private FromClause from;
    private WhereClause where;
    private GroupByClause groupBy;
    private HavingClause having;
    private OrderByClause orderBy;

    /**
     * Create a new <code>ProxyQueryImpl</code> instance.
     *
     * @param proxy
     *            the target proxy.
     * @param rootProxyClass
     *            the target proxy's entity class.
     */
    public AbstractProxyQueryImpl(
            T proxy,
            Class<?> rootProxyClass) {
        this.proxy = proxy;
        this.rootProxyClass = rootProxyClass;

        select = new SelectClause(false);
        this.from = new FromClause(rootProxyClass.getSimpleName(),
                proxy);
        this.where = new WhereClause();
        this.groupBy = new GroupByClause();
        this.having = new HavingClause();
        this.orderBy = new OrderByClause();
    }

    /**
     * {@inheritDoc}
     */
    public AbstractProxyQueryImpl<T> distinct() {

        setSelect(new SelectClause(getSelect(), true));

        return this;
    }

    /**
     * {@inheritDoc}
     */
    public AbstractProxyQueryImpl<T> groupBy(Object... selections) {

        List<Selection> selects = new ArrayList<Selection>();
        for (Object obj : selections) {
            selects.add(createQueryValue(obj));
        }
        groupBy = new GroupByClause(selects);

        return this;
    }

    /**
     * {@inheritDoc}
     */
    public AbstractProxyQueryImpl<T> orderBy(Object... sorts) {
        List<QuerySort> list = new ArrayList<QuerySort>();
        for (Object sortObject : sorts) {
            if (sortObject == null) {
                DirectionalQuerySort sort = new DirectionalQuerySort(
                        createQueryValue(sortObject), QuerySortOperator.ASC);
                list.add(sort);
            } else if (sortObject instanceof DirectionalQuerySort) {
                list.add((QuerySort) sortObject);
            }  else if (sortObject instanceof Selection) {
                list.add(new DirectionalQuerySort((Selection) sortObject, QuerySortOperator.ASC));
            }
        }
        orderBy = new OrderByClause(list);
        return this;
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public <S> TypedQuery<S> select(Class<S> constructorClass, Object... constructorArguments) {
        if (constructorClass == null) {
            throw new IllegalArgumentException("targetClass is required");
        }

        List<QueryValue> arugments = new ArrayList<QueryValue>(constructorArguments.length);
        for (Object object : constructorArguments) {
            arugments.add(createQueryValue(object));
        }

        return select(new ConstructorValue<S>(constructorClass, arugments));
    }

    /**
     * {@inheritDoc}
     */
    public <S> TypedQuery<S> select(S selection) {

        return select(createQueryValue(selection));
    }

    /**
     * {@inheritDoc}
     */
    public DynamicQuery select(Object... selections) {
        List<Selection> selects = new ArrayList<Selection>();
        for (Object obj : selections) {
            selects.add(createQueryValue(obj));
        }
        setSelect(new SelectClause(selects, getSelect().isDistinct()));

        setQuery(createDynamicQuery());

        return getQuery();
    }

    /**
     * Creates a new dynamic query (one that returns rows of type Object[]).
     *
     * @return a new dynamic query.
     */
    public abstract DynamicQuery createDynamicQuery();

    /**
     * {@inheritDoc}
     */
    public <S> RedirectingQueryValue<S> orWhere(S recordedMethodCall) {
        or();
        return new RedirectingQueryValue<S>(createQueryValue(recordedMethodCall), this);
    }

    /**
     * {@inheritDoc}
     */
    public <S> RedirectingQueryValue<S> andWhere(S recordedMethodCall) {
        return new RedirectingQueryValue<S>(createQueryValue(recordedMethodCall), this);
    }

    /**
     * {@inheritDoc}
     */
    public AbstractProxyQueryImpl<T> where(Conditional... conditionalExpressions) {
        where = new WhereClause(conditionalExpressions);
        return this;
    }

    /**
     * {@inheritDoc}
     */
    public AbstractProxyQueryImpl<T> having(Conditional... conditionalExpressions) {
        having = new HavingClause(conditionalExpressions);
        return this;
    }

    /**
     * {@inheritDoc}
     */
    public <S> Subquery<S> subquery(QueryValue<S> value) {

        List<Selection> list = new ArrayList<Selection>();
        list.add(value);

        setSelect(new SelectClause(list, getSelect().isDistinct()));

        return new SubqueryValue<S>(
                getSelect(),
                getFrom(),
                getWhere(),
                getGroupBy(),
                getHaving(),
                getOrderBy());
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public TypedQuery<T> select() {

        List<Selection> list = new ArrayList<Selection>();
        list.add(new SelectProxy(getRootProxy()));

        setSelect(new SelectClause(list, getSelect().isDistinct()));

        return createTypedQuery();
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public <S> TypedQuery<S> select(QueryValue<S> value) {

        List<Selection> list = new ArrayList<Selection>();
        list.add(value);

        setSelect(new SelectClause(list, getSelect().isDistinct()));

        return createTypedQuery();
    }

    @SuppressWarnings("unchecked")
    protected abstract TypedQuery createTypedQuery();

    /**
     * {@inheritDoc}
     */
    public SelectClause getSelect() {
        return select;
    }

    protected void setSelect(SelectClause select) {
        this.select = select;
    }

    /**
     * {@inheritDoc}
     */
    public FromClause getFrom() {
        return from;
    }

    /**
     * {@inheritDoc}
     */
    public WhereClause getWhere() {
        return where;
    }

    /**
     * {@inheritDoc}
     */
    public GroupByClause getGroupBy() {
        return groupBy;
    }

    /**
     * {@inheritDoc}
     */
    public HavingClause getHaving() {
        return having;
    }

    /**
     * {@inheritDoc}
     */
    public OrderByClause getOrderBy() {
        return orderBy;
    }

    /**
     * {@inheritDoc}
     */
    public T getRootProxy() {
        return proxy;
    }

    protected DynamicQuery getQuery() {
        return query;
    }

    protected void setQuery(DynamicQuery query) {
        this.query = query;
    }

    /**
     * {@inheritDoc}
     */
    public Class<?> getRootProxyClass() {
        return rootProxyClass;
    }

    /**
     * {@inheritDoc}
     */
    public DynamicQuery build() {
        if (query == null) {
            throw new IllegalStateException("select() was not called");
        }
        return query;
    }

    // TODO: unused?
    @SuppressWarnings("unchecked")
    <C> C join(
            Collection<C> collectionProperty,
            QueryJoinType joinType,
            QueryJoinModifier joinModifier) {

        RecordedMethodCall call = RecordingSessions.get().getSafeFirstCall();

        if (call.getParentProxy() == null) {
            throw new IllegalStateException("No parent-proxy for call: " + call);
        }

        QueryJoin parent = getFrom().findParentJoin(call.getParentProxy());

        if (parent == null) {
            throw new IllegalStateException("Could not find parent join for collection join: "
                    + call);
        }

        C joinProxy = (C) MethodCallUtils.proxy(call.getElementType(), "");

        QueryJoin join = new NestedPropertyJoin(
                parent,
                joinType,
                joinModifier,
                call.getName(),
                joinProxy);

        from = new FromClause(join, getFrom());

        return joinProxy;
    }

    /**
     * {@inheritDoc}
     */
    public <C> C join(Collection<C> collectionProperty) {
        return join(collectionProperty, QueryJoinType.INNER, QueryJoinModifier.NONE);
    }

    /**
     * {@inheritDoc}
     */
    public <C> C joinFetch(Collection<C> collectionProperty) {
        return join(collectionProperty, QueryJoinType.INNER, QueryJoinModifier.FETCH);
    }

    /**
     * {@inheritDoc}
     */
    public <C> C leftJoin(Collection<C> collectionProperty) {
        return join(collectionProperty, QueryJoinType.LEFT, QueryJoinModifier.NONE);
    }

    /**
     * {@inheritDoc}
     */
    public <C> C rightJoin(Collection<C> collectionProperty) {
        return join(collectionProperty, QueryJoinType.RIGHT, QueryJoinModifier.NONE);
    }

    @SuppressWarnings("unchecked")
    private QueryJoin findJoinParent(ProxyPathExpression targetProperty) {

        Object parentProxy = targetProperty.getCall().getParentProxy();

        if (parentProxy == null) {
            throw new IllegalStateException("No parent-proxy for call: "
                    + targetProperty.getCall());
        }
        if (targetProperty.getCall().getElementType() == null) {
            throw new IllegalStateException("No element-type for call: "
                    + targetProperty.getCall());
        }
        if (Collection.class.isAssignableFrom(targetProperty.getCall().getElementType())) {
            throw new IllegalStateException("Element-type for call must not be collection: "
                    + targetProperty.getCall());
        }

        QueryJoin parent = getFrom().findParentJoin(parentProxy);
        if (parent == null) {
            throw new IllegalArgumentException("Could not find parent join");
        }

        return parent;
    }

    @SuppressWarnings("unchecked")
    private <J> J join(
            QueryValue<J> targetQueryValue,
            QueryJoinType joinType,
            QueryJoinModifier joinModifier) {

        if (!(targetQueryValue instanceof ProxyPathExpression)) {
            throw new IllegalArgumentException("Invalid Join Argument, expected: "
                    + ProxyPathExpression.class.getSimpleName());
        }
        ProxyPathExpression targetProperty = (ProxyPathExpression) targetQueryValue;

        QueryJoin parent = findJoinParent(targetProperty);

        J joinProxy = (J) MethodCallUtils.proxy(
                targetProperty.getCall().getElementType(), "");

        QueryJoin join = new NestedPropertyJoin(
                parent,
                joinType,
                joinModifier,
                targetProperty.getCall().getName(),
                joinProxy);

        from = new FromClause(join, getFrom());

        return joinProxy;
    }

    /**
     * {@inheritDoc}
     */
    public <J> J join(QueryValue<J> targetProperty) {
        return join(targetProperty, QueryJoinType.INNER, QueryJoinModifier.NONE);
    }

    /**
     * {@inheritDoc}
     */
    public <J> J joinFetch(QueryValue<J> targetProperty) {
        return join(targetProperty, QueryJoinType.INNER, QueryJoinModifier.FETCH);
    }

    /**
     * {@inheritDoc}
     */
    public <J> J leftJoin(QueryValue<J> targetProperty) {
        return join(targetProperty, QueryJoinType.LEFT, QueryJoinModifier.NONE);
    }

    /**
     * {@inheritDoc}
     */
    public <J> J rightJoin(QueryValue<J> targetProperty) {
        return join(targetProperty, QueryJoinType.RIGHT, QueryJoinModifier.NONE);
    }

    /**
     * {@inheritDoc}
     */
    public <J> J thetaJoin(Object parentProxy, Class<J> joinClass) {
        if (joinClass == null) {
            throw new IllegalArgumentException("joinClass was not specified");
        }
        MethodCallUtils.validateIsProxy(parentProxy, "parentProxy");

        QueryJoin parent = getFrom().findParentJoin(parentProxy);
        if (parent == null) {
            throw new IllegalArgumentException("Could not find parent join");
        }

        J joinProxy = (J) MethodCallUtils.proxy(joinClass, "");

        ThetaJoin join = new ThetaJoin(
                parent,
                joinClass.getSimpleName(),
                joinProxy);

        from = new FromClause(join, getFrom());

        return joinProxy;
    }

    /**
     * {@inheritDoc}
     */
    public void processConditional(Conditional newConditional) {
        where = new WhereClause(where, newConditional);
    }
}
