/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.impl;

import java.util.HashMap;
import java.util.Map;

import org.cementframework.querybyproxy.shared.api.ProxyQueryBuilder;
import org.cementframework.querybyproxy.shared.api.ProxyQuerySession;
import org.cementframework.querybyproxy.shared.api.ProxyQuerySessions;
import org.cementframework.querybyproxy.shared.api.model.conditionals.ComparisonOperator;
import org.cementframework.querybyproxy.shared.api.model.conditionals.Conditional;
import org.cementframework.querybyproxy.shared.api.model.conditionals.GroupConditional;
import org.cementframework.querybyproxy.shared.api.model.conditionals.LogicGate;
import org.cementframework.querybyproxy.shared.api.model.conditionals.SubqueryConditional;
import org.cementframework.querybyproxy.shared.api.model.selections.AggregateSelection;
import org.cementframework.querybyproxy.shared.api.model.selections.QueryAggregate;
import org.cementframework.querybyproxy.shared.api.model.sorts.DirectionalQuerySort;
import org.cementframework.querybyproxy.shared.api.model.sorts.QuerySortOperator;
import org.cementframework.querybyproxy.shared.api.model.values.ProxyPathExpression;
import org.cementframework.querybyproxy.shared.api.model.values.ProxySelectionExpression;
import org.cementframework.querybyproxy.shared.api.model.values.QueryLiteral;
import org.cementframework.querybyproxy.shared.api.model.values.QueryParameter;
import org.cementframework.querybyproxy.shared.api.model.values.QueryValue;
import org.cementframework.querybyproxy.shared.api.model.values.SimpleQueryValue;
import org.cementframework.querybyproxy.shared.api.model.values.Subquery;
import org.cementframework.querybyproxy.shared.api.model.values.SubqueryModifierDecorator;
import org.cementframework.querybyproxy.shared.api.model.values.SubqueryOperator;
import org.cementframework.recordingproxy.api.RecordedMethodCall;
import org.cementframework.recordingproxy.api.RecordingSession;
import org.cementframework.recordingproxy.api.RecordingSessions;
import org.cementframework.recordingproxy.impl.MethodCallUtils;

/**
 * Retrieves proxy-method calls and creates query-model fragments.
 *
 * @author allenparslow
 */
class ProxyQueryBuilderImpl implements ProxyQueryBuilder {

    private Map<Class<?>, Object> proxyMap = new HashMap<Class<?>, Object>();

    @SuppressWarnings("unchecked")
    public <T> T createProxy(Class<T> type) {
        Object proxy = proxyMap.get(type);
        if (proxy == null) {
            proxy = MethodCallUtils.proxy(type);
            proxyMap.put(type, proxy);
        }
        return (T) proxy;
    }

    /**
     * {@inheritDoc}
     */
    public DirectionalQuerySort desc(Object instance) {

        return new DirectionalQuerySort(
                createQueryValue(instance),
                QuerySortOperator.DESC);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public QueryValue<Long> count(Object selection) {
        return (QueryValue) createAggregateSelection(selection, QueryAggregate.COUNT);
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> min(T selection) {
        return createAggregateSelection(selection, QueryAggregate.MIN);
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> max(T selection) {
        return createAggregateSelection(selection, QueryAggregate.MAX);
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> sum(T selection) {
        return createAggregateSelection(selection, QueryAggregate.SUM);
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> avg(T selection) {
        return createAggregateSelection(selection, QueryAggregate.AVG);
    }

    private <T> AggregateSelection<T> createAggregateSelection(
            T selection,
            QueryAggregate aggregate) {
        return new AggregateSelection<T>(
                aggregate,
                createQueryValue(selection, false));
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> count(SimpleQueryValue<T> selection) {
        return createAggregateSelectionSrict(selection, QueryAggregate.COUNT);
    }

    /**
     * {@inheritDoc}
     */
    public QueryValue<Long> count() {
        return createAggregateSelectionSrict(new QueryLiteral<Long>(1L),
                QueryAggregate.COUNT);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public QueryValue<Long> size(SimpleQueryValue selection) {
        QueryValue<Long> aggValue = (QueryValue<Long>) selection;
        return createAggregateSelectionSrict(aggValue, QueryAggregate.SIZE);
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> min(SimpleQueryValue<T> selection) {
        return createAggregateSelectionSrict(selection, QueryAggregate.MIN);
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> max(SimpleQueryValue<T> selection) {
        return createAggregateSelectionSrict(selection, QueryAggregate.MAX);
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> sum(SimpleQueryValue<T> selection) {
        return createAggregateSelectionSrict(selection, QueryAggregate.SUM);
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> avg(SimpleQueryValue<T> selection) {
        return createAggregateSelectionSrict(selection, QueryAggregate.AVG);
    }

    private <T> AggregateSelection<T> createAggregateSelectionSrict(
            SimpleQueryValue<T> selection,
            QueryAggregate aggregate) {
        return new AggregateSelection<T>(aggregate, selection);
    }

    /**
     * {@inheritDoc}
     */
    public Conditional or() {

        getSession().setNextLogicGate(LogicGate.OR);

        return null;
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public Conditional exists(Subquery subquery) {
        LogicGate logicGate = getSession().getNextLogicGate();

        return new SubqueryConditional(logicGate, ComparisonOperator.EXISTS, subquery);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public Conditional notExists(Subquery subquery) {
        LogicGate logicGate = getSession().getNextLogicGate();

        return new SubqueryConditional(logicGate, ComparisonOperator.NOT_EXISTS, subquery);
    }

    /**
     * {@inheritDoc}
     */
    public <T> Subquery<T> all(Subquery<T> subquery) {

        return new SubqueryModifierDecorator<T>(SubqueryOperator.ALL, subquery);
    }

    /**
     * {@inheritDoc}
     */
    public <T> Subquery<T> any(Subquery<T> subquery) {

        return new SubqueryModifierDecorator<T>(SubqueryOperator.ANY, subquery);
    }

    /**
     * {@inheritDoc}
     */
    public <T> Subquery<T> some(Subquery<T> subquery) {

        return new SubqueryModifierDecorator<T>(SubqueryOperator.SOME, subquery);
    }

    /**
     * {@inheritDoc}
     */
    public Conditional group(Conditional... conditionalExpressions) {

        LogicGate logicGate = getSession().getNextLogicGate();

        GroupConditional group = new GroupConditional(logicGate, conditionalExpressions);

        return group;
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public <T> QueryValue<T> get(Object instance, Class<T> returnType) {

        QueryValue<T> column;
        if (MethodCallUtils.isProxy(instance)) {
            RecordedMethodCall call = getRecordingSession().getLastCall();

            if (call != null) {
                column = new ProxyPathExpression(
                        call);
            } else {
                column = new ProxySelectionExpression(
                        instance);
            }
        } else {
            RecordedMethodCall call = getRecordingSession().getSafeLastCall();

            column = new ProxyPathExpression(
                    call);
        }

        return column;
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryValue<T> get(T instance) {

        return get(instance, null);
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public DirectionalQuerySort desc(SimpleQueryValue selection) {
        return new DirectionalQuerySort(selection, QuerySortOperator.DESC);
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryLiteral<T> literal(T target) {
        return new QueryLiteral<T>(target);
    }

    public <T> QueryValue<T> createQueryValue(T target) {
        return createQueryValue(target, true);
    }

    @SuppressWarnings("unchecked")
    public <T> QueryValue<T> createQueryValue(T target, boolean getFirst) {
        if (target != null && !MethodCallUtils.isProxy(target)) {
            if (target instanceof QueryValue) {
                return (QueryValue) target;
            } else {
                return new QueryLiteral<T>(target);
            }
        } else {
            return createProperty(target, getFirst);
        }
    }

    @SuppressWarnings("unchecked")
    private ProxyPathExpression createProperty(Object obj, boolean getFirst) {
        RecordedMethodCall call;
        RecordingSession recordingSession = RecordingSessions.get();
        if (getFirst) {
            call = recordingSession.getSafeFirstCall();
        } else {
            call = recordingSession.getSafeLastCall();
        }

        ProxyPathExpression selection = new ProxyPathExpression(call);

        return selection;
    }

    /**
     * {@inheritDoc}
     */
    public <T> QueryParameter<T> param(T target) {
        return new QueryParameter<T>(target);
    }

    private RecordingSession getRecordingSession() {
        return RecordingSessions.get();
    }

    private ProxyQuerySession getSession() {
        return ProxyQuerySessions.get();
    }
}
