/*
 * Copyright 2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cementframework.querybyproxy.shared.impl;

import org.cementframework.querybyproxy.shared.api.ProxyQueryBuilder;
import org.cementframework.querybyproxy.shared.api.model.conditionals.Conditional;
import org.cementframework.querybyproxy.shared.api.model.sorts.DirectionalQuerySort;
import org.cementframework.querybyproxy.shared.api.model.values.QueryValue;
import org.cementframework.querybyproxy.shared.api.model.values.Subquery;

/**
 * Static access to a query-builder.
 *
 * <p>Used for very terse query construction.</p>
 *
 * @author allenparslow
 */
public class StaticProxyQueryBuilder {
    private static final ProxyQueryBuilder INSTANCE = new ProxyQueryBuilderImpl();

    /**
     * Creates a "all" subquery modifier around a subquery.
     *
     * @param <T>
     *            type-inference.
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper.
     */
    public static <T> Subquery<T> all(Subquery<T> subquery) {
        return INSTANCE.all(subquery);
    }

    /**
     * Creates a "any" subquery modifier around a subquery.
     *
     * @param <T>
     *            type-inference.
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper.
     */
    public static <T> Subquery<T> any(Subquery<T> subquery) {
        return INSTANCE.any(subquery);
    }

    /**
     * Create a "avg" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> avg(QueryValue<T> selection) {
        return INSTANCE.avg(selection);
    }

    /**
     * Creates a avg aggregate using the specified recorded proxy method call.
     *
     * @param <T>
     *            type-inference.
     * @param proxyMethodCall
     *            a recorded proxy method call.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> avg(T proxyMethodCall) {
        return INSTANCE.avg(proxyMethodCall);
    }

    /**
     * Creates a "count(1)" expression.
     *
     * @return a query-value expression.
     */
    public static QueryValue<Long> count() {
        return INSTANCE.count();
    }

    /**
     * Create a "count" aggregate using the specified recorded proxy method
     * call.
     *
     * @param proxyMethodCall
     *            a recorded proxy method call. a recorded proxy method call.
     * @return a query-value expression.
     */
    public static QueryValue<Long> count(Object proxyMethodCall) {
        return INSTANCE.count(proxyMethodCall);
    }

    /**
     * Create a "count" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> count(QueryValue<T> selection) {
        return INSTANCE.count(selection);
    }

    /**
     * Creates proxy for the specified class.
     *
     * @param <T>
     *            type-inference.
     * @param type
     *            the type to proxy.
     * @return the new proxy.
     */
    public static <T> T createProxy(Class<T> type) {
        return INSTANCE.createProxy(type);
    }

    /**
     * Creates a query value expression used either a recorded method call, a
     * literal value or query parameter.
     *
     * @param <T>
     *            type-inference.
     * @param target
     *            the target value or recorded method call.
     * @param useFirstCall
     *            True, if the first call on the recording stack should be used.
     * @return the query value expression.
     */
    public static <T> QueryValue<T> createQueryValue(T target, boolean useFirstCall) {
        return INSTANCE.createQueryValue(target, useFirstCall);
    }

    /**
     * Creates a query value expression used either a recorded method call, a
     * literal value or query parameter.
     *
     * @param <T>
     *            type-inference.
     * @param target
     *            the target value or recorded method call.
     * @return the query value expression.
     */
    public static <T> QueryValue<T> createQueryValue(T target) {
        return INSTANCE.createQueryValue(target);
    }

    /**
     * Creates a query descending sort using the specified recorded proxy method
     * call.
     *
     * @param proxyMethodCall
     *            a recorded proxy method call.
     * @return a descending sort
     */
    public static DirectionalQuerySort desc(Object proxyMethodCall) {
        return INSTANCE.desc(proxyMethodCall);
    }

    /**
     * Creates a query descending sort using the specified selection.
     *
     * @param selection
     *            a selection.
     * @return a descending sort
     */
    @SuppressWarnings("unchecked")
    public static DirectionalQuerySort desc(QueryValue selection) {
        return INSTANCE.desc(selection);
    }

    /**
     * Creates a "exists" subquery modifier around a subquery.
     *
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper conditional-expression.
     */
    @SuppressWarnings("unchecked")
    public static Conditional exists(Subquery subquery) {
        return INSTANCE.exists(subquery);
    }

    /**
     * Gets a query-value expression with an overridden type using the last
     * recorded proxy method call.
     *
     * <p>
     * Commonly used when dealing with aliased properties as the alias type
     * might not be of the same type as the recorded method call's type
     * </p>
     *
     * @param <T>
     *            type-inference using the specified returnType.
     * @param proxyMethodCall
     *            a recorded proxy method call.
     * @param returnType
     *            The overridden return type.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> get(Object proxyMethodCall, Class<T> returnType) {
        return INSTANCE.get(proxyMethodCall, returnType);
    }

    /**
     * Gets a query-value expression using the last recorded proxy method call.
     *
     * @param <T>
     *            type-inference.
     * @param proxyMethodCall
     *            a recorded proxy method call.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> get(T proxyMethodCall) {
        return INSTANCE.get(proxyMethodCall);
    }

    /**
     * Creates a group "(...)" wrapper around the specified conditional
     * expressions.
     *
     * @param conditionalExpressions
     *            the conditional-expressions to group.
     * @return the group wrapper.
     */
    public static Conditional group(Conditional... conditionalExpressions) {
        return INSTANCE.group(conditionalExpressions);
    }

    /**
     * Produces a literal (non-parameter) expression using the passed object.
     *
     * <p>
     * SQL Injection Warning: Do NOT use with end-user specified text as it is
     * treated as a literal string during query compilation
     * </p>
     *
     * @param <T>
     *            type-inference.
     * @param object
     *            the target object to be used as a literal.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> literal(T object) {
        return INSTANCE.literal(object);
    }

    /**
     * Create a "max" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> max(QueryValue<T> selection) {
        return INSTANCE.max(selection);
    }

    /**
     * Creates a "max" aggregate using the specified recorded proxy method call.
     *
     * @param <T>
     *            type-inference.
     * @param proxyMethodCall
     *            a recorded proxy method call.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> max(T proxyMethodCall) {
        return INSTANCE.max(proxyMethodCall);
    }

    /**
     * Create a "min" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> min(QueryValue<T> selection) {
        return INSTANCE.min(selection);
    }

    /**
     * Creates a "min" aggregate using the specified recorded proxy method call.
     *
     * @param <T>
     *            type-inference.
     * @param proxyMethodCall
     *            a recorded proxy method call.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> min(T proxyMethodCall) {
        return INSTANCE.min(proxyMethodCall);
    }

    /**
     * Creates a "not exists" subquery modifier around a subquery.
     *
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper conditional-expression.
     */
    @SuppressWarnings("unchecked")
    public static Conditional notExists(Subquery subquery) {
        return INSTANCE.notExists(subquery);
    }

    /**
     * Used to specify that the next conditional-expression should be created as
     * an or-expression, overriding the default conjunction/logic-gate (which is
     * <b>and</b>).
     *
     * @return a conditional-expression placeholder.
     */
    public static Conditional or() {
        return INSTANCE.or();
    }

    /**
     * Produces a jpql bindable for the target object (produces ":paramX" syntax
     * during query compilation).
     *
     * @param <T>
     *            type-inference.
     * @param recordedMethodCall
     *            the recorded method call to be used a bindable query-parameter.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> param(T recordedMethodCall) {
        return INSTANCE.param(recordedMethodCall);
    }

    /**
     * Create a "size" modifier using the specified selection.
     *
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    @SuppressWarnings("unchecked")
    public static QueryValue<Long> size(QueryValue selection) {
        return INSTANCE.size(selection);
    }

    /**
     * Creates a "some" subquery modifier around a subquery.
     *
     * @param <T>
     *            type-inference.
     * @param subquery
     *            the subquery to wrap.
     * @return a modifier-subquery-wrapper.
     */
    public static <T> Subquery<T> some(Subquery<T> subquery) {
        return INSTANCE.some(subquery);
    }

    /**
     * Create a "sum" aggregate using the specified selection.
     *
     * @param <T>
     *            type-inference.
     * @param selection
     *            the expression to aggregate.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> sum(QueryValue<T> selection) {
        return INSTANCE.sum(selection);
    }

    /**
     * Creates a "sum" aggregate using the specified recorded proxy method call.
     *
     * @param <T>
     *            type-inference.
     * @param proxyMethodCall
     *            a recorded proxy method call.
     * @return a query-value expression.
     */
    public static <T> QueryValue<T> sum(T proxyMethodCall) {
        return INSTANCE.sum(proxyMethodCall);
    }
}
